// Copyright 2012 The Chromium Authors. All rights reserved.
// Use of this source code is governed by a BSD-style license that can be
// found in the LICENSE file.
/**
 * This software is contributed or developed by KYOCERA Corporation.
 * (C) 2019 KYOCERA Corporation
 */

package org.chromium.components.embedder_support.delegate;

import android.app.AlertDialog;
import android.app.Dialog;
import android.content.Context;
import android.content.DialogInterface;
import android.view.KeyEvent;
import android.view.LayoutInflater;
import android.view.View;
import android.widget.EditText;
import android.widget.Button;
import android.widget.TextView;
import android.widget.ScrollView;

import org.chromium.base.AfpConfigUtil;
import org.chromium.base.AfpUiInterfaceWrapper;

/**
 * UI for the color chooser that shows on the Android platform as a result of
 * &lt;input type=color &gt; form element.
 */
public class ColorPickerDialog extends AlertDialog implements OnColorChangedListener {
    private final ScrollView mScrollView;
    private final ColorPickerAdvanced mAdvancedColorPicker;

    private final ColorPickerSimple mSimpleColorPicker;

    private final Button mMoreButton;

    // The view up in the corner that shows the user the color they've currently selected.
    private final View mCurrentColorView;

    private final OnColorChangedListener mListener;

    private final int mInitialColor;

    private int mCurrentColor;

    private int mConfigValueSoftkey;
    private boolean mMoreButtonFocused;
    private AfpUiInterfaceWrapper mWrapper;

    /**
     * @param context The context the dialog is to run in.
     * @param listener The object to notify when the color is set.
     * @param color The initial color to set.
     * @param suggestions The list of suggestions.
     */
    public ColorPickerDialog(Context context, OnColorChangedListener listener, int color,
            ColorSuggestion[] suggestions) {
        super(context, 0);

        mListener = listener;
        mInitialColor = color;
        mCurrentColor = mInitialColor;
        mWrapper = new AfpUiInterfaceWrapper(context);
        mConfigValueSoftkey = AfpConfigUtil.getInstance().getConfigIntValue(AfpConfigUtil.KEY_PICKER_SOFTKEY);
        mMoreButtonFocused = false;

        // Initialize title
        LayoutInflater inflater =
                (LayoutInflater) context.getSystemService(Context.LAYOUT_INFLATER_SERVICE);

        // Initialize Set/Cancel buttons
        String positiveButtonText;
        if(mConfigValueSoftkey <= 1) {
            positiveButtonText = context.getString(R.string.color_picker_button_set);
        } else {
            positiveButtonText = context.getString(R.string.color_picker_button_cancel);
        }

        setButton(BUTTON_POSITIVE, positiveButtonText, new Dialog.OnClickListener() {
            @Override
            public void onClick(DialogInterface dialogInterface, int i) {
                if(mConfigValueSoftkey <= 1) {
                    tryNotifyColorSet(mCurrentColor);
                } else {
                    tryNotifyColorSet(mInitialColor);
                }
            }
        });

        // Note that with the color picker there's not really any such thing as
        // "cancelled".
        // The color picker flow only finishes when we return a color, so we
        // have to always
        // return something. The concept of "cancelled" in this case just means
        // returning
        // the color that we were initialized with.
        if(mConfigValueSoftkey <= 1) {
            setButton(BUTTON_NEGATIVE, mWrapper.getSystemString("webview_softkey_cancel"), new Dialog.OnClickListener() {
                @Override
                public void onClick(DialogInterface dialogInterface, int i) {
                    tryNotifyColorSet(mInitialColor);
                }
            });
        } else {
            setButton(BUTTON_NEUTRAL, mWrapper.getSystemString("webview_softkey_done"), new Dialog.OnClickListener() {
                @Override
                public void onClick(DialogInterface dialogInterface, int i) {
                    tryNotifyColorSet(mCurrentColor);
                }
            });
        }

        setOnCancelListener(new DialogInterface.OnCancelListener() {
            @Override
            public void onCancel(DialogInterface arg0) {
                tryNotifyColorSet(mInitialColor);
            }
        });

        mWrapper.assignButtonToSoftkey(this, true);
        if(!(mConfigValueSoftkey <= 1)) {
            setSoftkeyDoneAndSet(true);
        }

        // Initialize main content view
        View content = inflater.inflate(R.layout.color_picker_dialog_content, null);

        if(AfpConfigUtil.getInstance().getConfigBooleanValue(AfpConfigUtil.KEY_COLOR_PICKER_SCROLL)) {
            ScrollView scrollView = new ScrollView(context);
            setView(scrollView);
            scrollView.addView(content);
        } else {
            setView(content);
        }

        View title = content.findViewById(R.id.color_picker_dialog_title);
        mCurrentColorView = title.findViewById(R.id.selected_color_view);
        TextView titleText = (TextView) title.findViewById(R.id.title);
        titleText.setText(R.string.color_picker_dialog_title);

        // Initialize More button.
        mMoreButton = (Button) content.findViewById(R.id.more_colors_button);
        mMoreButton.setOnClickListener(new Button.OnClickListener() {
            @Override
            public void onClick(View v) {
                showAdvancedView();
            }
        });
        if(!(mConfigValueSoftkey <= 1)) {
            mMoreButton.setOnFocusChangeListener(new View.OnFocusChangeListener() {
                @Override
                public void onFocusChange(View v, boolean hasFocus) {
                    mMoreButtonFocused = hasFocus;
                }
            });
        }

        // Initialize advanced color view (hidden initially).
        mScrollView = (ScrollView) content.findViewById(R.id.color_picker_advanced_scr);
        mScrollView.setVisibility(View.GONE);
        mAdvancedColorPicker =
                (ColorPickerAdvanced) content.findViewById(R.id.color_picker_advanced);
        mAdvancedColorPicker.setVisibility(View.GONE);

        // Initialize simple color view (default view).
        mSimpleColorPicker = (ColorPickerSimple) content.findViewById(R.id.color_picker_simple);
        mSimpleColorPicker.init(suggestions, this);

        mSimpleColorPicker.setInitialFocus(mInitialColor);
        updateCurrentColor(mInitialColor);
    }

    /**
     * Listens to the ColorPicker for when the user has changed the selected color, and
     * updates the current color (the color shown in the title) accordingly.
     *
     * @param color The new color chosen by the user.
     */
    @Override
    public void onColorChanged(int color) {
        updateCurrentColor(color);
    }

    /**
     * Hides the simple view (the default) and shows the advanced one instead, hiding the
     * "More" button at the same time.
     */
    private void showAdvancedView() {
        // Only need to hide the borders, not the Views themselves, since the Views are
        // contained within the borders.
        View buttonBorder = findViewById(R.id.more_colors_button_border);
        buttonBorder.setVisibility(View.GONE);

        View simpleView = findViewById(R.id.color_picker_simple);
        simpleView.setVisibility(View.GONE);

        mScrollView.setVisibility(View.VISIBLE);
        mAdvancedColorPicker.setVisibility(View.VISIBLE);
        mAdvancedColorPicker.setListener(this);
        mAdvancedColorPicker.setColor(mCurrentColor);
        mAdvancedColorPicker.focusSeekBar();
        setSoftkeyDoneAndSet(false);
    }

    /**
     * Tries to notify any listeners that the color has been set.
     */
    private void tryNotifyColorSet(int color) {
        if (mListener != null) mListener.onColorChanged(color);
    }

    /**
     * Updates the internal cache of the currently selected color, updating the colorful little
     * box in the title at the same time.
     */
    private void updateCurrentColor(int color) {
        mCurrentColor = color;
        if (mCurrentColorView != null) mCurrentColorView.setBackgroundColor(color);
    }

    private void setSoftkeyDoneAndSet(boolean showSet) {
        if(mConfigValueSoftkey <= 1) {
            mWrapper.setDialogSoftkey(this, AfpUiInterfaceWrapper.SET_BRANK_STR, AfpUiInterfaceWrapper.LOCAL_RES_ID_NONE, AfpUiInterfaceWrapper.INDEX_CSK);
        } else {
            if (showSet) {
                mWrapper.setDialogSoftkey(this, "", R.string.color_picker_button_set, AfpUiInterfaceWrapper.INDEX_SK2);
            } else {
                mWrapper.setDialogSoftkey(this, AfpUiInterfaceWrapper.SET_BRANK_STR, AfpUiInterfaceWrapper.LOCAL_RES_ID_NONE, AfpUiInterfaceWrapper.INDEX_SK2);
            }
        }
    }

    @Override
    public boolean dispatchKeyEvent(KeyEvent event) {
        if(!(mConfigValueSoftkey <= 1)) {
            final int keyCode = event.getKeyCode();
            if (event.getAction() == KeyEvent.ACTION_UP) {
                switch (keyCode) {
                    case KeyEvent.KEYCODE_DPAD_CENTER:
                    case KeyEvent.KEYCODE_ENTER:
                        getButton(DialogInterface.BUTTON_NEUTRAL).performClick();
                        return true;
                    case KeyEvent.KEYCODE_SOFT_RIGHT:
                        if (mAdvancedColorPicker.getVisibility() == View.VISIBLE) {
                            return false;
                        }
                        if (mMoreButtonFocused) {
                            showAdvancedView();
                        } else {
                            int position = mSimpleColorPicker.getSelectedItemPosition();
                            if (position < 0) {
                                return false;
                            }
                            ColorSuggestion suggestion = (ColorSuggestion) mSimpleColorPicker.getAdapter().getItem(position);
                            updateCurrentColor(suggestion.mColor);
                        }
                        return true;
                    default:
                        break;
                }
            }
        }
        return super.dispatchKeyEvent(event);
    }
}
