// Copyright 2019 The Chromium Authors. All rights reserved.
// Use of this source code is governed by a BSD-style license that can be
// found in the LICENSE file.
/**
 * This software is contributed or developed by KYOCERA Corporation.
 * (C) 2019 KYOCERA Corporation
 */

#include "third_party/blink/renderer/core/page/spatial_navigation_controller.h"

#include "third_party/blink/public/platform/web_scroll_into_view_params.h"
#include "third_party/blink/renderer/core/dom/document.h"
#include "third_party/blink/renderer/core/dom/element.h"
#include "third_party/blink/renderer/core/dom/element_traversal.h"
#include "third_party/blink/renderer/core/dom/node.h"
#include "third_party/blink/renderer/core/events/keyboard_event.h"
#include "third_party/blink/renderer/core/frame/local_frame.h"
#include "third_party/blink/renderer/core/frame/settings.h"
#include "third_party/blink/renderer/core/frame/visual_viewport.h"
#include "third_party/blink/renderer/core/html/html_area_element.h"
#include "third_party/blink/renderer/core/html/html_frame_owner_element.h"
#include "third_party/blink/renderer/core/html/html_image_element.h"
#include "third_party/blink/renderer/core/input/event_handler.h"
#include "third_party/blink/renderer/core/layout/hit_test_location.h"
#include "third_party/blink/renderer/core/layout/hit_test_result.h"
#include "third_party/blink/renderer/core/layout/layout_object.h"
#include "third_party/blink/renderer/core/page/focus_controller.h"
#include "third_party/blink/renderer/core/page/page.h"
#include "third_party/blink/renderer/core/page/spatial_navigation.h"
#include "third_party/blink/renderer/platform/geometry/layout_rect.h"

#include "third_party/blink/renderer/core/css/style_change_reason.h"

namespace blink {

namespace {

SpatialNavigationDirection FocusDirectionForKey(KeyboardEvent* event) {
  if (event->ctrlKey() || event->metaKey() || event->shiftKey())
    return SpatialNavigationDirection::kNone;

  SpatialNavigationDirection ret_val = SpatialNavigationDirection::kNone;
  if (event->key() == "ArrowDown")
    ret_val = SpatialNavigationDirection::kDown;
  else if (event->key() == "ArrowUp")
    ret_val = SpatialNavigationDirection::kUp;
  else if (event->key() == "ArrowLeft")
    ret_val = SpatialNavigationDirection::kLeft;
  else if (event->key() == "ArrowRight")
    ret_val = SpatialNavigationDirection::kRight;

  // TODO(bokan): We should probably assert that we don't get anything else but
  // currently KeyboardEventManager sends non-arrow keys here.

  return ret_val;
}

void ClearFocusInExitedFrames(LocalFrame* old_frame,
                              const LocalFrame* const new_frame) {
  while (old_frame && new_frame != old_frame) {
    // Focus is going away from this document, so clear the focused element.
    old_frame->GetDocument()->ClearFocusedElement();
    old_frame->GetDocument()->SetSequentialFocusNavigationStartingPoint(
        nullptr);
    Frame* parent = old_frame->Tree().Parent();
    old_frame = DynamicTo<LocalFrame>(parent);
  }
}

static bool AreElementOfInvisibility(const FocusCandidate& candidate)
{
  LayoutPoint point(candidate.rect_in_root_frame.Center());
  LayoutSize padding(candidate.rect_in_root_frame.Size().Width() / 2, candidate.rect_in_root_frame.Size().Height() / 2);

  Frame* frame = candidate.visible_node->GetDocument().GetFrame();
  if (!frame)
    return true;

  if (!frame->IsMainFrame()) {
    LayoutRect r = candidate.visible_node->Node::BoundingBox();
    point = r.Center();
  } else {
    frame = candidate.visible_node->GetDocument().GetPage()->MainFrame();
  }

  HitTestLocation location(HitTestLocation(point), padding);
  HitTestResult result =
      To<LocalFrame>(frame)->GetEventHandler().HitTestResultAtLocation(location,
          HitTestRequest::kReadOnly | HitTestRequest::kActive |
          HitTestRequest::kIgnoreClipping | HitTestRequest::kListBased |
          HitTestRequest::kPenetratingList);

  HitTestResult::NodeSet nodeList = result.ListBasedTestResult();
  HitTestResult::NodeSet::const_iterator last = nodeList.end();
  for (HitTestResult::NodeSet::const_iterator it = nodeList.begin(); it != last; ++it) {
    if (candidate.visible_node->contains(it->Get())) {
      return false;
    } else if(candidate.visible_node->IsFrameOwnerElement() &&
          FrameOwnerElement(candidate)->ContentFrame() == it->Get()->GetDocument().GetFrame()){
        return false;
    }
  }
  return true;
}

static bool CheckFocusCandidateOfInterSection(const FocusCandidate& current_interest,
                                                const FocusCandidate& candidate,
                                                CandidateDistances distance,
                                                int *alignment,
                                                FocusCandidate* best_candidate,
                                                CandidateDistances* best_distance) {
  bool isInner = false;

  if (candidate.is_offscreen || candidate.rect_in_root_frame.IsEmpty())
    return false;

  if (CheckIfInterSectionElement(current_interest, candidate, &isInner)) {
    if (best_candidate->IsNull()) {
      if (isInner)
          *alignment = RectsAlignment::Full;
      return true;
    }
    if (distance.navigation_axis_distance < best_distance->navigation_axis_distance) {
      if (isInner)
          *alignment = RectsAlignment::Full;
      return true;
    }
  }

  return false;
}

bool IsNearlyFullScreenSizeNode(Node* visible_node, SpatialNavigationDirection direction, bool deflate) {

  LayoutRect rect_in_root_frame;
  if (auto* area = ToHTMLAreaElementOrNull(*visible_node)) {
    HTMLImageElement* image = area->ImageElement();
    if (!image || !image->GetLayoutObject())
      return false;

    rect_in_root_frame = StartEdgeForAreaElement(*area, direction);
  } else {
    if (!visible_node->GetLayoutObject())
      return false;

    rect_in_root_frame = NodeRectInRootFrame(visible_node, true /* ignore border */);
  }

  if(visible_node->IsElementNode()) {
    LayoutSize viewSize = LayoutSize(To<LocalFrame>(visible_node->GetDocument().GetPage()->MainFrame())->View()->FrameRect().Size());

    if(deflate) {
      viewSize.SetWidth(viewSize.Width() - 2);
      viewSize.SetHeight(viewSize.Height() - 2);
    }

    if(!GetNodeZIndex(visible_node))
      return false;

    if(EVisibility::kVisible != visible_node->GetLayoutObject()->Style()->Visibility())
      return false;

    if(EDisplay::kNone == visible_node->GetLayoutObject()->Style()->Display())
      return false;

    if(!visible_node->GetLayoutObject()->Style()->Opacity() || !GetOpacity(visible_node))
      return false;

    if(!visible_node->GetLayoutObject()->Style()->HasBackground())
        return false;

    if( viewSize.Width().ToDouble() * 0.1 <=  rect_in_root_frame.Width().ToDouble()
        && viewSize.Height().ToDouble() * 0.1 <= rect_in_root_frame.Height().ToDouble())
        return true;
  }
    return false;
}

static void UpdateFocusCandidateForOverflow(SpatialNavigationDirection direction,
                                             const FocusCandidate& current_interest,
                                             const FocusCandidate& candidate,
                                             FocusCandidate* best_candidate,
                                             CandidateDistances* best_distance,
                                             int* best_alignment,
                                             std::unordered_map<std::string, FocusCandidate> popupElements) {
  if(candidate.rect_in_root_frame.Contains(current_interest.rect_in_root_frame))
    return;

  int candidateZindex = GetNodeZIndex(candidate.visible_node);
  for (std::unordered_map<std::string, FocusCandidate>::iterator it = popupElements.begin(); it != popupElements.end(); ++it) {
    FocusCandidate popupElement = it->second;
    if(candidateZindex < GetNodeZIndex(popupElement.visible_node)) {
      if(popupElement.rect_in_root_frame.Contains(candidate.rect_in_root_frame)) {
        return;
      }
    }
  }

  switch (direction) {
    case SpatialNavigationDirection::kUp:
      if(current_interest.rect_in_root_frame.MaxY() < candidate.rect_in_root_frame.MaxY())
        return;

      if(current_interest.rect_in_root_frame.MaxY() == candidate.rect_in_root_frame.MaxY())
        if(current_interest.rect_in_root_frame.X() > candidate.rect_in_root_frame.X() ||
          current_interest.rect_in_root_frame.MaxX() < candidate.rect_in_root_frame.MaxX())
          return;

      if(!best_candidate->IsNull())
        if(best_candidate->rect_in_root_frame.MaxY() > candidate.rect_in_root_frame.MaxY())
          return;
      break;

    case SpatialNavigationDirection::kDown:
      if(current_interest.rect_in_root_frame.Y() > candidate.rect_in_root_frame.Y())
        return;

      if(current_interest.rect_in_root_frame.Y() == candidate.rect_in_root_frame.Y())
        if(current_interest.rect_in_root_frame.X() > candidate.rect_in_root_frame.X() ||
            current_interest.rect_in_root_frame.MaxX() < candidate.rect_in_root_frame.MaxX())
          return;

      if(!best_candidate->IsNull())
        if(best_candidate->rect_in_root_frame.Y() < candidate.rect_in_root_frame.Y())
          return;
      break;

    case SpatialNavigationDirection::kLeft:
      if(current_interest.rect_in_root_frame.MaxX() < candidate.rect_in_root_frame.MaxX())
        return;

      if(current_interest.rect_in_root_frame.MaxX() == candidate.rect_in_root_frame.MaxX())
        if(current_interest.rect_in_root_frame.Y() > candidate.rect_in_root_frame.Y() ||
            current_interest.rect_in_root_frame.MaxY() < candidate.rect_in_root_frame.MaxY())
          return;

      if(!best_candidate->IsNull())
        if(best_candidate->rect_in_root_frame.MaxX() > candidate.rect_in_root_frame.MaxX())
          return;
      break;

    case SpatialNavigationDirection::kRight:
      if(current_interest.rect_in_root_frame.X() > candidate.rect_in_root_frame.X())
        return;

      if(current_interest.rect_in_root_frame.X() == candidate.rect_in_root_frame.X())
        if(current_interest.rect_in_root_frame.Y() > candidate.rect_in_root_frame.Y() ||
            current_interest.rect_in_root_frame.MaxY() < candidate.rect_in_root_frame.MaxY() )
          return;

      if(!best_candidate->IsNull())
        if(best_candidate->rect_in_root_frame.X() < candidate.rect_in_root_frame.X())
          return;
      break;

    default:
      break;
  }

  if(AreElementOfInvisibility(candidate))
    return;

  CandidateDistances distances;
  distances.distance = 0.0;
  distances.navigation_axis_distance = 0.0;
  distances.orthogonal_axis_distance = 0.0;
  *best_candidate = candidate;
  *best_distance = distances;
  *best_alignment = RectsAlignment::Full;
}

static void UpdateFocusCandidateForDirection(SpatialNavigationDirection direction,
                                               const FocusCandidate& current_interest,
                                               const FocusCandidate& candidate,
                                               CandidateDistances distance,
                                               int alignment,
                                               FocusCandidate* best_candidate,
                                               CandidateDistances* best_distance,
                                               int* best_alignment) {

  if (!CheckFocusCandidateOfInterSection(current_interest, candidate, distance, &alignment, best_candidate, best_distance)) {
    switch (direction) {
      case SpatialNavigationDirection::kLeft:
      case SpatialNavigationDirection::kRight:
        if (alignment > *best_alignment)
          break;

        if (alignment == *best_alignment) {
          if (distance.navigation_axis_distance < best_distance->navigation_axis_distance)
            break;

          if (distance.navigation_axis_distance == best_distance->navigation_axis_distance)
            if (distance.orthogonal_axis_distance < best_distance->orthogonal_axis_distance)
              break;
        }
        return;
      case SpatialNavigationDirection::kUp:
      case SpatialNavigationDirection::kDown:
        if (distance.distance < best_distance->distance)
          break;

        if (distance.distance == best_distance->distance)
          if (alignment > *best_alignment)
            break;

        return;
      default:
        return;
    }
  }

  *best_candidate = candidate;
  *best_distance = distance;
  *best_alignment = alignment;
}

// Determines whether the given candidate is closer to the current interested
// node (in the given direction) than the current best. If so, it'll replace
// the current best.
static void ConsiderForBestCandidate(SpatialNavigationDirection direction,
                                     const FocusCandidate& current_interest,
                                     const FocusCandidate& candidate,
                                     FocusCandidate* best_candidate,
                                     CandidateDistances* best_distance,
                                     int* best_alignment,
                                     std::unordered_map<std::string, FocusCandidate> popupElements) {
  DCHECK(candidate.visible_node->IsElementNode());
  DCHECK(candidate.visible_node->GetLayoutObject());

  // Ignore iframes that don't have a src attribute
  if (FrameOwnerElement(candidate) &&
      (!FrameOwnerElement(candidate)->ContentFrame() ||
       candidate.rect_in_root_frame.IsEmpty()))
    return;

  if (IsSameLine(direction, current_interest, candidate)) {
    CandidateDistances distances;
    distances.distance = 0.0;
    distances.navigation_axis_distance = 0.0;
    distances.orthogonal_axis_distance = 0.0;
    *best_candidate = candidate;
    *best_distance = distances;
    *best_alignment = RectsAlignment::Full;
    return;
  } else if (AreElementsOnSameLine(current_interest, candidate)) {
    return;
  }

  if (GetOpacity(candidate.visible_node) <= 0.0f) {
    if(!DeprecatedEqualIgnoringCase(candidate.visible_node->nodeName(),"a") &&
       !DeprecatedEqualIgnoringCase(candidate.visible_node->nodeName(),"button") &&
       !DeprecatedEqualIgnoringCase(candidate.visible_node->nodeName(),"input") &&
       !DeprecatedEqualIgnoringCase(candidate.visible_node->nodeName(),"textarea") &&
       !DeprecatedEqualIgnoringCase(candidate.visible_node->nodeName(),"select")) {
      return;
    }
  }

  if((current_interest.rect_in_root_frame.Contains(candidate.rect_in_root_frame) ||
        current_interest.rect_in_root_frame.Intersects(candidate.rect_in_root_frame)) &&
      (current_interest.rect_in_root_frame != candidate.rect_in_root_frame) &&
      (current_interest.visible_node && current_interest.visible_node->GetDocument().GetFrame()->IsMainFrame() &&
        candidate.visible_node && !candidate.visible_node->GetLayoutObject()->IsLayoutIFrame())) {
    UpdateFocusCandidateForOverflow(direction, current_interest, candidate, best_candidate, best_distance, best_alignment, popupElements);
    return;
  }

  CandidateDistances distance =
      ComputeDistanceDataForNode(direction, current_interest, candidate);
  if (distance.distance == MaxDistance())
    return;

  int alignment =
      AlignmentForRects(direction, current_interest.rect_in_root_frame, candidate.rect_in_root_frame,
          LayoutSize(To<LocalFrame>(candidate.visible_node->GetDocument().GetPage()->MainFrame())->View()->GetScrollableArea()->VisibleContentRect().Size()));

  int candidateZindex = GetNodeZIndex(candidate.visible_node);
  for (std::unordered_map<std::string, FocusCandidate>::iterator it = popupElements.begin(); it != popupElements.end(); ++it) {
    FocusCandidate popupElement = it->second;
    if(candidateZindex < GetNodeZIndex(popupElement.visible_node)){
      if(popupElement.rect_in_root_frame.Contains(candidate.rect_in_root_frame)){
        return;
      }
    }
  }

  if (best_candidate->IsNull()) {
    CheckFocusCandidateOfInterSection(current_interest, candidate, distance, &alignment, best_candidate, best_distance);
    *best_candidate = candidate;
    *best_distance = distance;
    *best_alignment = alignment;
    return;
  }

  LayoutRect intersection_rect = Intersection(
      candidate.rect_in_root_frame, best_candidate->rect_in_root_frame);
  if (!intersection_rect.IsEmpty() &&
      !AreElementsOnSameLine(*best_candidate, candidate) &&
      intersection_rect == candidate.rect_in_root_frame) {
    // If 2 nodes are intersecting, do hit test to find which node in on top.
    LayoutUnit x = intersection_rect.X() + intersection_rect.Width() / 2;
    LayoutUnit y = intersection_rect.Y() + intersection_rect.Height() / 2;
    if (!IsA<LocalFrame>(
            candidate.visible_node->GetDocument().GetPage()->MainFrame()))
      return;
    HitTestLocation location(IntPoint(x.ToInt(), y.ToInt()));
    HitTestResult result =
        candidate.visible_node->GetDocument()
            .GetPage()
            ->DeprecatedLocalMainFrame()
            ->GetEventHandler()
            .HitTestResultAtLocation(
                location, HitTestRequest::kReadOnly | HitTestRequest::kActive |
                              HitTestRequest::kIgnoreClipping);
    if(0 <= best_distance->distance && best_distance->distance <= distance.distance)
      return;
    if (candidate.visible_node->contains(result.InnerNode())) {
      *best_candidate = candidate;
      *best_distance = distance;
      *best_alignment = alignment;
      return;
    }
    if (best_candidate->visible_node->contains(result.InnerNode()))
      return;
  }

  if (best_candidate->rect_in_root_frame == candidate.rect_in_root_frame) {
    if (candidate.visible_node->IsDescendantOf(best_candidate->visible_node)) {
      if (AreElementOfInvisibility(candidate))
        return;
      *best_candidate = candidate;
      *best_distance = distance;
      *best_alignment = alignment;
      return;
    }
  }

  UpdateFocusCandidateForDirection(direction, current_interest, candidate, distance, alignment, best_candidate, best_distance, best_alignment);
}

static void AddPopupElement(FocusCandidate& candidate, int candidateZindex, std::unordered_map<std::string, FocusCandidate>& popupElements) {
  std::string zIndex = std::to_string(candidateZindex);
  std::string rectX = std::to_string(candidate.rect_in_root_frame.X().ToInt());
  std::string rectY = std::to_string(candidate.rect_in_root_frame.Y().ToInt());
  std::string rectMaxX = std::to_string(candidate.rect_in_root_frame.MaxX().ToInt());
  std::string rectMaxY = std::to_string(candidate.rect_in_root_frame.MaxY().ToInt());
  popupElements["z" + zIndex + "x" + rectX + "y" + rectY + "mx" + rectMaxX + "my" + rectMaxY] = candidate;
}

}  // namespace

// static
SpatialNavigationController* SpatialNavigationController::Create(Page& page) {
  return MakeGarbageCollected<SpatialNavigationController>(page);
}

SpatialNavigationController::SpatialNavigationController(Page& page)
    : page_(&page) {}

bool SpatialNavigationController::HandleArrowKeyboardEvent(
    KeyboardEvent* event) {
  DCHECK(page_->GetSettings().GetSpatialNavigationEnabled());

  SpatialNavigationDirection direction = FocusDirectionForKey(event);
  if (direction == SpatialNavigationDirection::kNone)
    return false;

  return Advance(direction);
}

bool SpatialNavigationController::HandleEnterKeyboardEvent(
    KeyboardEvent* event) {
  DCHECK(page_->GetSettings().GetSpatialNavigationEnabled());

  if (interest_element_) {
    interest_element_->focus(FocusParams(SelectionBehaviorOnFocus::kReset,
                                         kWebFocusTypeSpatialNavigation,
                                         nullptr));
  }

  return true;
}

bool SpatialNavigationController::HandleEscapeKeyboardEvent(
    KeyboardEvent* event) {
  DCHECK(page_->GetSettings().GetSpatialNavigationEnabled());

  if (!interest_element_)
    return false;

  if (interest_element_->IsFocusedElementInDocument())
    interest_element_->blur();
  else
    MoveInterestTo(nullptr);

  return true;
}

Element* SpatialNavigationController::GetInterestedElement() const {
  if (RuntimeEnabledFeatures::FocuslessSpatialNavigationEnabled())
    return interest_element_;

  Frame* frame = page_->GetFocusController().FocusedOrMainFrame();
  auto* local_frame = DynamicTo<LocalFrame>(frame);
  if (!local_frame)
    return nullptr;

  Document* document = local_frame->GetDocument();
  if (!document)
    return nullptr;

  return document->ActiveElement();
}

void SpatialNavigationController::DidDetachFrameView() {
  // If the interested element's view was lost (frame detached, navigated,
  // etc.) then reset navigation.
  if (interest_element_ && !interest_element_->GetDocument().View())
    interest_element_ = nullptr;
}

void SpatialNavigationController::Trace(blink::Visitor* visitor) {
  visitor->Trace(interest_element_);
  visitor->Trace(page_);
}

bool SpatialNavigationController::Advance(
    SpatialNavigationDirection direction) {
  Node* interest_node = StartingNode();
  if (!interest_node)
    return false;

  interest_node->GetDocument().UpdateStyleAndLayoutIgnorePendingStylesheets();

  Node* container = ScrollableAreaOrDocumentOf(interest_node);

  const LayoutRect visible_rect(page_->GetVisualViewport().VisibleRect());
  const LayoutRect start_box =
      SearchOrigin(visible_rect, interest_node, direction);

  if (IsScrollableAreaOrDocument(interest_node) &&
      (!IsOffscreen(interest_node) || !container)) {
    // A visible scroller has interest. Search inside of it from one of its
    // edges.
    LayoutRect edge = OppositeEdge(direction, start_box);
    if (AdvanceWithinContainer(*interest_node, edge, direction, nullptr))
      return true;
  }

  // The interested scroller had nothing. Let's search outside of it.
  Node* skipped_tree = interest_node;
  while (container) {
    if (AdvanceWithinContainer(*container, start_box, direction, skipped_tree))
      return true;

    // Containers are not focused “on the way out”. This prevents containers
    // from acting as “focus traps”. Take <c> <a> </c> <b>. Focus can move from
    // <a> to <b> but not from <a> to the scroll container <c>. If we'd allow
    // focus to move from <a> to <c>, the user would never be able to exit <c>.
    // When the scroll container <c> is focused, we move focus back to <a>...
    skipped_tree = container;
    // Nothing found in |container| so search the parent container.
    container = ScrollableAreaOrDocumentOf(container);

    // TODO(bokan): This needs to update the parent document when the _current_
    // container is a document since we're crossing the document boundary.
    // Currently this will fail if we're going from an inner document to a
    // sub-scroller in a parent document.
    if (auto* document = DynamicTo<Document>(container))
      document->UpdateStyleAndLayoutIgnorePendingStylesheets();
  }

  return false;
}

FocusCandidate SpatialNavigationController::FindNextCandidateInContainer(
    Node& container,
    const LayoutRect& starting_rect_in_root_frame,
    SpatialNavigationDirection direction,
    Node* interest_child_in_container) {
  Element* element = ElementTraversal::FirstWithin(container);

  FocusCandidate current_interest;
  current_interest.rect_in_root_frame = starting_rect_in_root_frame;
  current_interest.focusable_node = interest_child_in_container;
  current_interest.visible_node = interest_child_in_container;

  FocusCandidate best_candidate;
  CandidateDistances best_distance;
  int best_alignment = RectsAlignment::None;

  std::vector<FocusCandidate> onScreenElements;
  std::unordered_map<std::string, FocusCandidate> popupElements;

  DeflateFocusCandidate(current_interest);

  for (; element;
       element =
           IsScrollableAreaOrDocument(element)
               ? ElementTraversal::NextSkippingChildren(*element, &container)
               : ElementTraversal::Next(*element, &container)) {

    Node* visible_node = nullptr;
    if (auto* area = ToHTMLAreaElementOrNull(*element)) {
      HTMLImageElement* image = area->ImageElement();
      if (!image || !image->GetLayoutObject())
        continue;

      visible_node = image;
    } else {
      if (!element->GetLayoutObject())
        continue;

      visible_node = element;
    }

    if (IsOffscreen(visible_node))
      continue;

    if (!IsValidCandidate(*element)) {
      if(IsNearlyFullScreenSizeNode(visible_node, direction, true)) {
        FocusCandidate candidate = FocusCandidate(element, direction);
        if (candidate.IsNull())
          continue;
        DeflateFocusCandidate(candidate);
        AddPopupElement(candidate, candidate.visible_node->GetLayoutObject()->Style()->ZIndex(), popupElements);
      }
      continue;
    }

    if (HasRemoteFrame(element))
      continue;

    if(EVisibility::kVisible != visible_node->GetLayoutObject()->Style()->Visibility())
      continue;

    if(EDisplay::kNone == visible_node->GetLayoutObject()->Style()->Display())
      continue;

    if(!visible_node->GetLayoutObject()->Style()->Opacity()) {
      if(!DeprecatedEqualIgnoringCase(visible_node->nodeName(),"a") &&
         !DeprecatedEqualIgnoringCase(visible_node->nodeName(),"button") &&
         !DeprecatedEqualIgnoringCase(visible_node->nodeName(),"input") &&
         !DeprecatedEqualIgnoringCase(visible_node->nodeName(),"textarea") &&
         !DeprecatedEqualIgnoringCase(visible_node->nodeName(),"select"))
        continue;
    }

    FocusCandidate candidate = FocusCandidate(element, direction);
    if (candidate.IsNull())
      continue;

    DeflateFocusCandidate(candidate);

    if (element == interest_child_in_container) {
      AddPopupElement(candidate, visible_node->GetLayoutObject()->Style()->ZIndex(), popupElements);
      continue;
    }

    if(IsNearlyFullScreenSizeNode(candidate.visible_node, direction, false)) {
      AddPopupElement(candidate, candidate.visible_node->GetLayoutObject()->Style()->ZIndex(), popupElements);
    }

    onScreenElements.push_back(candidate);
  }

  for(std::vector<FocusCandidate>::iterator it = onScreenElements.begin(); it != onScreenElements.end(); ++it) {
    FocusCandidate candidate = *it;
    ConsiderForBestCandidate(direction, current_interest, candidate,
                             &best_candidate, &best_distance, &best_alignment, popupElements);
    if(IsSameLine(direction, current_interest, best_candidate))
      break;
  }

  return best_candidate;
}

bool SpatialNavigationController::AdvanceWithinContainer(
    Node& container,
    const LayoutRect& starting_rect_in_root_frame,
    SpatialNavigationDirection direction,
    Node* interest_child_in_container) {
  return AdvanceWithinContainer(container, starting_rect_in_root_frame, direction, interest_child_in_container, nullptr);
}

bool SpatialNavigationController::AdvanceWithinContainer(
    Node& container,
    const LayoutRect& starting_rect_in_root_frame,
    SpatialNavigationDirection direction,
    Node* interest_child_in_container,
    Node* last_candidate_node) {
  DCHECK(IsScrollableAreaOrDocument(&container));

  FocusCandidate candidate =
      FindNextCandidateInContainer(container, starting_rect_in_root_frame,
                                   direction, interest_child_in_container);

  if (candidate.IsNull()) {
    if (last_candidate_node)
      return false;

    // Nothing to focus in this container, scroll if possible.
    // NOTE: If no scrolling is performed (i.e. ScrollInDirection returns
    // false), the spatial navigation algorithm will skip this container.
    if (ScrollInDirection(&container, direction)) {
      if (container.IsDocumentNode()) {
        Document* document = DynamicTo<Document>(container);

        if(document && SpatialNavigationDirection::kDown == direction){
          document->SpeechElementAutoNext();
        } else if (document && SpatialNavigationDirection::kUp == direction) {
          document->ClearReachedLastElement();
        }
      }else if (IsHTMLTextAreaElement(ToElement(container))){
        To<LocalFrame>(page_->GetFocusController().FocusedOrMainFrame())->GetDocument()->SpeechElementAutoNext();
      }
      return true;
    } else {
      return false;
    }
  }

  if(candidate.is_offscreen) {
    ScrollInDirection(&container, direction);
    return true;
  }

  if(FrameOwnerElement(candidate)) {
    if (AdvanceWithinContainer(*To<LocalFrame>(FrameOwnerElement(candidate)->ContentFrame())->GetDocument(), starting_rect_in_root_frame, direction, interest_child_in_container)) {
      return true;
    } else {
        return AdvanceWithinContainer(container, NodeRectInRootFrame(candidate.focusable_node, true /* ignore border */), direction, interest_child_in_container, nullptr);
    }
  }

  if(IsScrollableAreaOrDocument(candidate.focusable_node))
    if (AdvanceWithinContainer(*candidate.focusable_node, starting_rect_in_root_frame, direction, interest_child_in_container, candidate.focusable_node))
      return true;

  Element* element = ToElement(candidate.focusable_node);
  DCHECK(element);
  MoveInterestTo(element);
  return true;
}

Node* SpatialNavigationController::StartingNode() {
  if (RuntimeEnabledFeatures::FocuslessSpatialNavigationEnabled()) {
    if (interest_element_ && interest_element_->isConnected() &&
        interest_element_->GetDocument().GetFrame()) {
      // If an iframe is interested, start the search from its document node.
      // This matches the behavior in the focus case below where focusing a
      // frame means the focused document doesn't have a focused element and so
      // we return the document itself.
      if (auto* frame_owner =
              DynamicTo<HTMLFrameOwnerElement>(interest_element_.Get()))
        return frame_owner->contentDocument();

      return interest_element_;
    }

    if (auto* main_local_frame = DynamicTo<LocalFrame>(page_->MainFrame()))
      return main_local_frame->GetDocument();

    return nullptr;
  }

  // FIXME: Directional focus changes don't yet work with RemoteFrames.
  const auto* current_frame =
      DynamicTo<LocalFrame>(page_->GetFocusController().FocusedOrMainFrame());
  if (!current_frame)
    return nullptr;

  Document* focused_document = current_frame->GetDocument();
  if (!focused_document)
    return nullptr;

  Node* focused_element = focused_document->FocusedElement();
  if (!focused_element)  // An iframe's document is focused.
    focused_element = focused_document;

  return focused_element;
}

void SpatialNavigationController::MoveInterestTo(Node* next_node) {
  DCHECK(!next_node || next_node->IsElementNode());
  Element* element = ToElement(next_node);

  if (RuntimeEnabledFeatures::FocuslessSpatialNavigationEnabled()) {
    if (interest_element_) {
      interest_element_->blur();
      interest_element_->SetNeedsStyleRecalc(
          kLocalStyleChange, StyleChangeReasonForTracing::Create(
                                 style_change_reason::kPseudoClass));
    }

    interest_element_ = element;

    if (interest_element_) {
      interest_element_->SetNeedsStyleRecalc(
          kLocalStyleChange, StyleChangeReasonForTracing::Create(
                                 style_change_reason::kPseudoClass));

      LayoutObject* layout_object = interest_element_->GetLayoutObject();
      DCHECK(layout_object);

      layout_object->ScrollRectToVisible(
          element->BoundingBoxForScrollIntoView(), WebScrollIntoViewParams());
    }
    return;
  }

  if (!element)
    return;

  // Before focusing the new element, check if we're leaving an iframe (= moving
  // focus out of an iframe). In this case, we want the exited [nested] iframes
  // to lose focus. This is tested in snav-iframe-nested.html.
  LocalFrame* old_frame = page_->GetFocusController().FocusedFrame();
  ClearFocusInExitedFrames(old_frame, next_node->GetDocument().GetFrame());

  element->focus(FocusParams(SelectionBehaviorOnFocus::kReset,
                             kWebFocusTypeSpatialNavigation, nullptr));
}

bool SpatialNavigationController::IsValidCandidate(
    const Element& element) const {
  return element.IsKeyboardFocusable();
}

}  // namespace blink
